# io.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>
# Code adapted from Alex Tseng, Avanti Shrikumar, and Ziga Avsec

import numpy
import torch
import pandas

import pyfaidx
import pyBigWig

from tqdm import tqdm

from .utils import one_hot_encode
from .utils import characters

from memelite.io import read_meme as memelite_read_meme


def _interleave_loci(loci, chroms=None):
	"""An internal function for processing the provided loci.

	There are two aspects that have to be considered when processing the loci.
	The first is that the user can pass in either strings containing filenames
	or pandas DataFrames. The second is that the user can pass in a single
	value or a list of values, and if a list of values the resulting dataframes
	must be interleaved.

	If a set of chromosomes is provided, each dataframe will be filtered to
	loci on those chromosomes before interleaving. If a more complicated form
	of filtering is desired, one should pre-filter the dataframes and pass those
	into this function for interleaving.


	Parameters
	----------
	loci: str, pandas.DataFrame, or list of those
		A filename to load, a pandas DataFrame in bed-format, or a list of
		either.

	chroms: list or None, optional
		A set of chromosomes to restrict the loci to. This is done before
		interleaving to ensure balance across sets of loci. If None, do not
		do filtering. Default is None.


	Returns
	-------
	interleaved_loci: pandas.DataFrame
		A single pandas DataFrame that interleaves rows from each of the
		provided examples.
	"""

	if chroms is not None:
		if not isinstance(chroms, (list, tuple)):
			raise ValueError("Provided chroms must be a list.")

	if isinstance(loci, (str, pandas.DataFrame)):
		loci = [loci]
	elif not isinstance(loci, (list, tuple)):
		raise ValueError("Provided loci must be a string or pandas " +
			"DataFrame, or a list/tuple of those.")

	names = ['chrom', 'start', 'end']
	loci_dfs = []
	for i, df in enumerate(loci):
		if isinstance(df, str):
			df = pandas.read_csv(df, sep='\t', usecols=[0, 1, 2], 
				header=None, index_col=False, names=names)
			
		elif isinstance(df, pandas.DataFrame):
			df = df.iloc[:, [0, 1, 2]].copy()
		
		else:
			raise ValueError("Provided loci must be a string or pandas " +
				"DataFrame, or a list/tuple of those.")

		if chroms is not None:
			df = df[numpy.isin(df['chrom'], chroms)]

		df['idx'] = numpy.arange(len(df)) * len(loci) + i
		loci_dfs.append(df)

	loci = pandas.concat(loci_dfs)
	loci = loci.set_index("idx").sort_index().reset_index(drop=True)
	return loci


def _load_signals(signals):
	"""An internal function for loading signals.

	The passed in signals must be a list but can either be a list of strings,
	which are interpreted as strings for bigwig files that should be opened,
	or dictionaries where the keys are chromosome names and the values are
	numpy arrays of values across the chromosome, which are kept as is.


	Parameters
	----------
	signals: list of strings or dicts or None
		A list of strings for bigwig files or dictionaries of numpy arrays.


	Returns
	-------
	_signals: list of dicts
		A list of either pointers to opened bigwig files or dictionaries of
		numpy a
	"""

	if signals is None:
		return None

	_signals = []
	for i, signal in enumerate(signals):
		if isinstance(signal, str):
			signal = pyBigWig.open(signal)
		elif not isinstance(signal, dict):
			raise ValueError("Signals must either be a list of strings " +
				"or a list of dictionaries.")
		elif not isinstance(list(signal.values())[0], numpy.ndarray):
			raise ValueError("Values in dictionaries must be numpy.ndarrays.")

		_signals.append(signal)

	return _signals


def _extract_locus_signal(signals, chrom, start, end):
	"""An internal function for extracting signal from a single locus.

	This function takes in a set of signals and a single locus and extracts
	the signal from each one of the loci.


	Parameters
	----------
	signals: list of pyBigWig objects or dictionaries
		A list of opened pyBigWig objects or dictionaries where the keys are
		chromosomes and the values are the signal at each position in the
		chromosome.

	chrom: str
		The name of the chromosome. Must be a key in the signals.

	start: int
		The starting coordinate to extract from, inclusive and base-0.

	end: int
		The ending coordinate to extract from, exclusive and base-0.


	Returns
	-------
	values: list of numpy.ndarrays, shape=(len(signals), end-start)
		The extracted signal from each of the signal files.
	"""

	if not isinstance(signals, (list, tuple)):
		raise ValueError("Provided signals must be in the form of a list.")

	values = []
	for i, signal in enumerate(signals):
		if isinstance(signal, dict):
			values_ = signal[chrom][start:end]
		else:
			try:
				values_ = signal.values(chrom, start, end, numpy=True)
			except:
				print(f"Warning: {chrom} {start} {end} not " +
					"valid bigwig indexes. Using zeros instead.")
				values_ = numpy.zeros(end-start)

		values_ = numpy.nan_to_num(values_)
		values.append(values_)

	return values


def extract_loci(loci, sequences, signals=None, in_signals=None, chroms=None, 
	in_window=2114, out_window=1000, max_jitter=0, min_counts=None,
	max_counts=None, target_idx=0, n_loci=None, alphabet=['A', 'C', 'G', 'T'], 
	ignore=['N'], return_filtered=False, verbose=False):
	"""Extract sequence and signal information for each provided locus.

	This function will take in a set of loci, sequences, and optionally signals,
	and return the sequences and signals at each of the loci. Each of these
	parameters can be a filename, which is loaded internally, or an appropriate
	Python object (see below for details). The nomenclature `in/out` refer to
	he expected inputs and outputs of the downstream machine learning model, 
	not this function.

	For each locus a sequence window of size `in_window` will be extracted from
	the sequences file and each of the `input_signals` files if provided, and
	a window of size `out_window` will be extracted from each of the `signals`
	files if provided. These windows are centered at the middle of the provided
	regions but all be of the same size, regardless of the size of the peak.

	If `max_jitter` is provided, it will expand the windows for both the input
	and output. The results are not actually jittered, but this expanded window
	allows for downstream data generators to created jittered data while
	reducing the memory footprint of the returned data.

	There are a few reasons that the returned elements may not match one-to-one
	with the provided loci:

		- (1) If any of the coordinates fall off the end of chromosomes after
		accounting for jitter, the locus will be removed.

		- (2) If any of the loci fall on chromosomes not in a provided list,
		they will be removed.

		- (3) If min_counts or max_counts are specified and the locus has a 
		number of counts not in those boundaries.
 

	Parameters
	----------
	loci: str or pandas.DataFrame or list/tuple of such
		Either the path to a bed file or a pandas DataFrame object containing
		three columns: the chromosome, the start, and the end, of each locus
		to train on. Alternatively, a list or tuple of strings/DataFrames where
		the intention is to train on the interleaved concatenation, i.e., when
		you want to train on peaks and negatives.

	sequences: str or dictionary
		Either the path to a fasta file to read from or a dictionary where the
		keys are the unique set of chromosoms and the values are one-hot
		encoded sequences as numpy arrays or memory maps.

	signals: list of strs or list of dictionaries or None, optional
		A list of filepaths to bigwig files, where each filepath will be read
		using pyBigWig, or a list of dictionaries where the keys are the same
		set of unique chromosomes and the values are numpy arrays or memory
		maps. If None, no signal tensor is returned. Default is None.

	input_signals: list of strs or list of dictionaries or None, optional
		A list of filepaths to bigwig files, where each filepath will be read
		using pyBigWig, or a list of dictionaries where the keys are the same
		set of unique chromosomes and the values are numpy arrays or memory
		maps. If None, no tensor is returned. Default is None. 

	chroms: list or None, optional
		A set of chromosomes to extact loci from. Loci in other chromosomes
		in the locus file are ignored. If None, all loci are used. Default is
		None.

	in_window: int, optional
		The input window size. Default is 2114.

	out_window: int, optional
		The output window size. Default is 1000.

	max_jitter: int, optional
		The maximum amount of jitter to add, in either direction, to the
		midpoints that are passed in. Default is 0.

	min_counts: float or None, optional
		The minimum number of counts, summed across the length of each example
		and across all tasks, needed to be kept. If None, no minimum. Default 
		is None.

	max_counts: float or None, optional
		The maximum number of counts, summed across the length of each example
		and across all tasks, needed to be kept. If None, no maximum. Default 
		is None.  

	target_idx: int, optional
		When specifying `min_counts` or `max_counts`, the single `signal`
		file to use when determining if a region has a number of counts in
		that range. Default is 0.

	n_loci: int or None, optional
		A cap on the number of loci to return. Note that this is not the
		number of loci that are considered. The difference is that some
		loci may be filtered out for various reasons, and those are not
		counted towards the total. If None, no cap. Default is None.

	alphabet : set or tuple or list
		A pre-defined alphabet where the ordering of the symbols is the same
		as the index into the returned tensor, i.e., for the alphabet ['A', 'B']
		the returned tensor will have a 1 at index 0 if the character was 'A'.
		Characters outside the alphabet are ignored and none of the indexes are
		set to 1. Default is ['A', 'C', 'G', 'T'].

	ignore: list, optional
		A list of characters to ignore in the sequence, meaning that no bits
		are set to 1 in the returned one-hot encoding. Put another way, the
		sum across characters is equal to 1 for all positions except those
		where the original sequence is in this list. Default is ['N'].
		
	return_filtered: bool, optional
		Whether to return a tensor containing whether each element in the provided
		loci have been filtered out because of falling off the edge of chromosomes
		or the signal not falling in the specified boundaries. Default is False.

	verbose: bool, optional
		Whether to display a progress bar while loading. Default is False.


	Returns
	-------
	seqs: torch.tensor, shape=(n, 4, in_window+2*max_jitter)
		The extracted sequences in the same order as the loci in the locus
		file after optional filtering by chromosome.

	signals: torch.tensor, shape=(n, len(signals), out_window+2*max_jitter)
		The extracted signals where the first dimension is in the same order
		as loci in the locus file after optional filtering by chromosome and
		the second dimension is in the same order as the list of signal files.
		If no signal files are given, this is not returned.

	in_signals: torch.tensor, shape=(n, len(in_signals),out_window+2*max_jitter)
		The extracted in signals where the first dimension is in the same order
		as loci in the locus file after optional filtering by chromosome and
		the second dimension is in the same order as the list of in signal files.
		If no in signal files are given, this is not returned.
	
	filtered: torch.tensor, shape=(n,), dtype=bool
		Whether each locus in the provided list has been kept (True) or filtered
		(False) due to the coordinates falling off the edges of the loci. Only
		returned if return_filtered=True.
	"""

	seqs, signals_, in_signals_ = [], [], []
	filtered = []
	in_width, out_width = in_window // 2, out_window // 2
	if signals is None and in_signals is None:
		out_width = 0

	# Load the sequences
	loci = _interleave_loci(loci, chroms)

	if isinstance(sequences, str):
		sequences = pyfaidx.Fasta(sequences)

	# Load the signal and optional in signal tracks if filenames are given
	signals = _load_signals(signals)
	in_signals = _load_signals(in_signals)

	desc = "Loading Loci"
	d = not verbose

	max_width = max(in_width, out_width)
	for chrom, start, end in tqdm(loci.values, disable=d, desc=desc):
		mid = start + (end - start) // 2
		start = mid - max(out_width, in_width) - max_jitter
		end = mid + max(out_width, in_width) + max_jitter

		if isinstance(sequences, dict):
			chrom_length = sequences[chrom].shape[-1]
		else:
			chrom_length = len(sequences[chrom])

		if start < 0 or end >= chrom_length:
			filtered.append(False)
			continue

		if n_loci is not None and len(seqs) == n_loci:
			break 

		# Extract a window of signal using the output size
		start = mid - out_width - max_jitter
		end = mid + out_width + max_jitter + (out_window % 2)

		if signals is not None:
			signal = _extract_locus_signal(signals, chrom, start, end)

			if min_counts is not None and signal[target_idx].sum() < min_counts:
				filtered.append(False)
				continue

			if max_counts is not None and signal[target_idx].sum() > max_counts:
				filtered.append(False)
				continue

			signals_.append(signal)

		# Extract a window of signal using the input size
		start = mid - in_width - max_jitter
		end = mid + in_width + max_jitter + (in_window % 2)

		if in_signals is not None:
			in_signal = _extract_locus_signal(in_signals, chrom, start, end)
			in_signals_.append(in_signal)

		# Extract a window of sequence using the input size
		if isinstance(sequences, dict):
			seq = sequences[chrom][:, start:end]
		else:
			seq = one_hot_encode(sequences[chrom][start:end].seq.upper(),
				alphabet=alphabet, ignore=ignore)

		filtered.append(True)
		seqs.append(seq)

	if not isinstance(sequences, dict):
		sequences.close()

	seqs = torch.from_numpy(numpy.stack(seqs))
	y_return = [seqs]

	if signals is not None:
		y_return.append(torch.from_numpy(numpy.stack(signals_)))

	if in_signals is not None:
		y_return.append(torch.from_numpy(numpy.stack(in_signals_)))
		
	if return_filtered:
		filtered = torch.tensor(filtered)
		y_return.append(torch.where(filtered == True)[0])

	return y_return[0] if len(y_return) == 1 else y_return


def one_hot_to_fasta(X, filename, mode='w', headers=None, 
	alphabet=['A', 'C', 'G', 'T']):
	"""Write out one-hot encoded sequences to a FASTA file.
	
	This function will take a set of one-hot encoded sequences and convert them to
	characters and write them out in FASTA format. If headers are provided for
	each sequence, these are used, otherwise the numeric index is used.
	"""
	
	with open(filename, mode=mode) as outfile:
		for i, X_seq in enumerate(X):
			X_chars = characters(X_seq, alphabet=alphabet)
			
			if headers is None:
				outfile.write("> {}\n".format(i))
			else:
				outfile.write("> {}\n".format(headers[i]))
			
			for start in range(0, len(X_chars), 80):
				outfile.write(X_chars[start:start+80] + "\n")
				
			outfile.write("\n")


def read_meme(filename, n_motifs=None):
	"""Read a MEME file and return a dictionary of PWMs.

	This method takes in the filename of a MEME-formatted file to read in
	and returns a dictionary of the PWMs where the keys are the metadata
	line and the values are the PWMs.

	This function is a wrapper around the memelite one, except that it returns
	torch tensors instead of numpy arrays.


	Parameters
	----------
	filename: str
		The filename of the MEME-formatted file to read in


	Returns
	-------
	motifs: dict
		A dictionary of the motifs in the MEME file.
	"""

	motifs = memelite_read_meme(filename, n_motifs=n_motifs)
	motifs = {name: torch.from_numpy(pwm) for name, pwm in motifs.items()}
	return motifs


def read_vcf(filename):
	"""Read a VCF file into a pandas DataFrame

	This function takes in the name of a file that is VCF formatted and returns
	a pandas DataFrame with the comments filtered out. This will only return the
	columns that are most commonly provided in VCF files.


	Parameters
	----------
	filename: str


	Returns
	-------
	vcf: pandas.DataFrame
		A pandas DataFrame containing the rows.
	"""

	names = ["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", 
		"FORMAT"]
	dtypes = {name: str for name in names}
	dtypes['POS'] = int

	vcf = pandas.read_csv(filename, delimiter='\t', comment='#', names=names, 
		dtype=dtypes, usecols=range(9))
	return vcf


# test_utils.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pandas


from tangermeme.utils import _validate_input
from tangermeme.utils import characters
from tangermeme.utils import one_hot_encode
from tangermeme.utils import reverse_complement
from tangermeme.utils import random_one_hot
from tangermeme.utils import chunk
from tangermeme.utils import unchunk
from tangermeme.utils import extract_signal

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal



def test_validate_input_shape():
	X = torch.randn(1)
	assert_raises(ValueError, _validate_input, X, "X", shape=(2, 1))
	assert_raises(ValueError, _validate_input, X, "X", shape=(2,))
	assert_raises(ValueError, _validate_input, X, "X", shape=(-1, -1))
	assert_raises(ValueError, _validate_input, X, "X", shape=(-1, 1))
	_validate_input(X, "X", shape=(1,))
	_validate_input(X, "X", shape=(-1,))

	X = torch.randn(5, 3)
	assert_raises(ValueError, _validate_input, X, "X", shape=(-1,))
	assert_raises(ValueError, _validate_input, X, "X", shape=(3,))
	assert_raises(ValueError, _validate_input, X, "X", shape=(5,))
	assert_raises(ValueError, _validate_input, X, "X", shape=(1, 4))
	assert_raises(ValueError, _validate_input, X, "X", shape=(-1, -1, -1))
	assert_raises(ValueError, _validate_input, X, "X", shape=(5, 3, -1))
	_validate_input(X, "X", shape=(5, 3))
	_validate_input(X, "X", shape=(-1, 3))
	_validate_input(X, "X", shape=(5, -1))
	_validate_input(X, "X", shape=(-1, -1))


def test_validate_input_dtype():
	X = torch.randn(5, 3)
	assert_raises(ValueError, _validate_input, X, "X", dtype=torch.float16)
	assert_raises(ValueError, _validate_input, X, "X", dtype=torch.float64)
	assert_raises(ValueError, _validate_input, X, "X", dtype=torch.int8)
	assert_raises(ValueError, _validate_input, X, "X", dtype=torch.int32)
	_validate_input(X, "X", dtype=torch.float32)


def test_validate_input_min_value():
	X = torch.randn(100, 100)
	assert_raises(ValueError, _validate_input, X, "X", min_value=0.0)
	_validate_input(torch.abs(X), "X", min_value=0.0)

	X[X < 0] = 0
	_validate_input(X, "X", min_value=0.0)

	X = X.type(torch.int8)
	_validate_input(X, "X", min_value=0.0)


def test_validate_input_max_value():
	X = torch.randn(100, 100)
	assert_raises(ValueError, _validate_input, X, "X", max_value=0.0)
	_validate_input(-torch.abs(X), "X", max_value=0.0)

	X[X > 0] = 0
	_validate_input(X, "X", max_value=0.0)

	X = X.type(torch.int8)
	_validate_input(X, "X", max_value=0.0)


def test_validate_input_ohe():
	X = random_one_hot((1, 4, 10), random_state=0)

	_validate_input(X, "X", ohe=True, ohe_dim=1)
	_validate_input(X.type(torch.float64), "X", ohe=True, ohe_dim=1)
	_validate_input(X.type(torch.int8), "X", ohe=True, ohe_dim=1)
	_validate_input(X.type(torch.int32), "X", ohe=True, ohe_dim=1)
	_validate_input(X.permute(1, 0, 2), "X", ohe=True, ohe_dim=0)
	_validate_input(X.permute(2, 0, 1), "X", ohe=True, ohe_dim=2)

	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=0)
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=2)
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=-1)
	assert_raises(ValueError, _validate_input, X + 0.2, "X", ohe=True, 
		ohe_dim=1)
	assert_raises(ValueError, _validate_input, X - 1, "X", ohe=True, 
		ohe_dim=1)

	X[0, 2, 3] = 1
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1)

	X[0, :, 3] = 0
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1)

	X = random_one_hot((1, 4, 10), random_state=0)[0]
	_validate_input(X, "X", ohe=True, ohe_dim=0)
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1)

	X = random_one_hot((1, 4, 10), random_state=0)[0, :, 0]
	_validate_input(X, "X", ohe=True, ohe_dim=0)
	assert_raises(IndexError, _validate_input, X, "X", ohe=True, ohe_dim=1)


def test_validate_input_allow_N():
	X = random_one_hot((2, 4, 10), random_state=0)
	_validate_input(X, "X", ohe=True, ohe_dim=1)

	X[0, :, 0] = 0
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1)
	_validate_input(X, "X", ohe=True, ohe_dim=1, allow_N=True)

	X[0, 1, 0] = 2
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1,
		allow_N=True)

	X[0, :, 0] = 1
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1,
		allow_N=True)

	X[0, 0, 0] = 0
	X[0, 1, 0] = 1
	X[0, 2, 0] = -1
	X[0, 3, 0] = 1
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1,
		allow_N=True)

	X = random_one_hot((2, 4, 10), random_state=0).float()
	_validate_input(X, "X", ohe=True, ohe_dim=1, allow_N=True)

	X[0, 1, 0] = 0.5
	assert_raises(ValueError, _validate_input, X, "X", ohe=True, ohe_dim=1,
		allow_N=True)


##


def test_characters_ohe():
	seq = 'GCTAC'
	ohe = torch.tensor([
		[0, 0, 0, 1, 0],
		[0, 1, 0, 0, 1],
		[1, 0, 0, 0, 0],
		[0, 0, 1, 0, 0]
	])

	seq_chars = characters(ohe)

	assert isinstance(seq_chars, str)
	assert len(seq_chars) == 5
	assert seq_chars == seq


def test_character_pwm():
	seq = 'GCTAC'
	ohe = torch.tensor([
		[0.25, 0.00, 0.10, 0.95, 0.00],
		[0.20, 1.00, 1.00, 0.05, 1.00],
		[0.30, 0.00, 0.30, 0.00, 0.00],
		[0.25, 0.00, 3.00, 0.00, 0.00]
	])

	seq_chars = characters(ohe)

	assert isinstance(seq_chars, str)
	assert len(seq_chars) == 5
	assert seq_chars == seq


def test_characters_alphabet():
	seq = 'GCTAC'
	ohe = torch.tensor([
		[0.25, 0.00, 0.10, 0.95, 0.00],
		[0.20, 1.00, 1.00, 0.05, 1.00],
		[0.30, 0.00, 0.30, 0.00, 0.00],
		[0.25, 0.00, 3.00, 0.00, 0.00],
		[0.00, 0.00, 0.00, 0.00, 0.00]
	])

	seq_chars = characters(ohe, ['A', 'C', 'G', 'T', 'N'])

	assert isinstance(seq_chars, str)
	assert len(seq_chars) == 5
	assert seq_chars == seq


def test_characters_raise_alphabet():
	seq = 'GCTAC'
	ohe = torch.tensor([
		[0.25, 0.00, 0.10, 0.95, 0.00],
		[0.20, 1.00, 1.00, 0.05, 1.00],
		[0.30, 0.00, 0.30, 0.00, 0.00],
		[0.25, 0.00, 3.00, 0.00, 0.00]
	])

	assert_raises(ValueError, characters, ohe, ['A', 'C', 'G'])
	assert_raises(ValueError, characters, ohe, ['A', 'C', 'G', 'T', 'N'])


def test_characters_raise_dimensions():
	seq = 'GCTAC'
	#this will work for shape (1,4,5) but not for (N,4,5) where N > 1
	ohe = torch.tensor([[
		[0.25, 0.00, 0.10, 0.95, 0.00],
		[0.20, 1.00, 1.00, 0.05, 1.00],
		[0.30, 0.00, 0.30, 0.00, 0.00],
		[0.25, 0.00, 3.00, 0.00, 0.00]
	]])
	
	assert characters(ohe) == seq
	
	ohe = torch.concat([ohe, ohe], dim=0)
	assert_raises(ValueError, characters, ohe, ['A', 'C', 'G', 'T'])
	
	ohe = torch.tensor([0.25, 0.00, 0.10, 0.95, 0.00])
	assert_raises(ValueError, characters, ohe, ['A', 'C', 'G', 'T'])


def test_characters_raise_ties():
	seq = 'GCTAC'
	ohe = torch.tensor([
		[0.25, 0.00, 0.10, 0.95, 0.00],
		[0.20, 1.00, 1.00, 0.05, 1.00],
		[0.30, 1.00, 0.30, 0.00, 0.00],
		[0.25, 0.00, 3.00, 0.00, 0.00]
	])

	assert_raises(ValueError, characters, ohe, ['A', 'C', 'G', 'T'])
	assert characters(ohe, force=True) == seq


##


def test_one_hot_encode():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0]
	])
	seq_ohe = one_hot_encode(seq)

	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (4, 5)
	assert torch.all(seq_ohe == ohe)

	seq = 'CCGTC'
	ohe = torch.tensor([
		[0, 0, 0, 0, 0],
		[1, 1, 0, 0, 1],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0]
	])
	seq_ohe = one_hot_encode(seq)

	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (4, 5)
	assert torch.all(seq_ohe == ohe)


	seq = 'AAAAA'
	ohe = torch.tensor([
		[1, 1, 1, 1, 1],
		[0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0]
	])
	seq_ohe = one_hot_encode(seq)

	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (4, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_N():
	seq = 'ACGNNTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 0, 0, 1],
		[0, 1, 0, 0, 0, 0, 0],
		[0, 0, 1, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 1, 0]
	])
	seq_ohe = one_hot_encode(seq)

	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (4, 7)
	assert torch.all(seq_ohe == ohe)

	seq = 'NNNNN'
	ohe = torch.tensor([
		[0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0]
	])
	seq_ohe = one_hot_encode(seq)

	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (4, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_dtype():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0]
	])
	seq_ohe = one_hot_encode(seq, dtype=torch.float32)

	assert seq_ohe.dtype == torch.float32
	assert seq_ohe.shape == (4, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_alphabet():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0],
		[0, 0, 0, 0, 0]
	])
	seq_ohe = one_hot_encode(seq, alphabet=['A', 'C', 'G', 'T', 'Z'])
	
	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (5, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_ignore():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0],
		[0, 0, 0, 0, 0]
	])
	seq_ohe = one_hot_encode(seq, alphabet=['A', 'C', 'G', 'T', 'N'], ignore=[])
	
	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (5, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_alphabet():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0],
		[0, 0, 0, 0, 0]
	])
	seq_ohe = one_hot_encode(seq, alphabet=['A', 'C', 'G', 'T', 'Z'])
	
	assert seq_ohe.dtype == torch.int8
	assert seq_ohe.shape == (5, 5)
	assert torch.all(seq_ohe == ohe)


def test_one_hot_encode_raises_alphabet():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0]
	])

	assert_raises(ValueError, one_hot_encode, seq, ['A', 'C', 'G'])


def test_one_hot_encode_raises_ignore():
	seq = 'ACGTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0],
		[0, 0, 0, 0, 0]
	])

	assert_raises(ValueError, one_hot_encode, seq, ['A', 'C', 'G', 'T', 'N'])


def test_one_hot_encode_lower_raises():
	seq = 'AcgTA'
	ohe = torch.tensor([
		[1, 0, 0, 0, 1],
		[0, 1, 0, 0, 0],
		[0, 0, 1, 0, 0],
		[0, 0, 0, 1, 0]
	])

	assert_raises(ValueError, one_hot_encode, seq)


###


def test_reverse_complement():
	assert 'TTGACC' == reverse_complement('GGTCAA')
	assert 'TTTTTT' == reverse_complement('AAAAAA')
	assert 'ACGT' == reverse_complement('ACGT')

	f = one_hot_encode
	assert_array_almost_equal(f('TTGACC'), reverse_complement(f('GGTCAA')))
	assert_array_almost_equal(f('TTTTTT'), reverse_complement(f('AAAAAA')))
	assert_array_almost_equal(f('ACGT'), reverse_complement(f('ACGT')))


def test_reverse_complement_mapping():
	assert 'AAAAAA' == reverse_complement('GGTCAA', 
		complement_map={'A': 'A', 'C': 'A', 'G': 'A', 'T': 'A'})

	assert 'TTGACC' == reverse_complement('GGTCaa',
		complement_map={'a': 'T', 'C': 'G', 'G': 'C', 'T': 'A'})

	f = one_hot_encode
	assert_array_almost_equal(f('TTGACC'), reverse_complement(f('GGTCAA')))
	assert_raises(AssertionError, assert_array_almost_equal, f('TTGACC'), 
		reverse_complement(f('GGTCAA'), complement_map={'A': 'T', 'G': 'C', 
			'T': 'A', 'C': 'G'}))


def test_reverse_reverse_complement():
	rc = reverse_complement
	f = one_hot_encode

	assert 'TTGACC' == rc(rc('TTGACC'))
	assert 'TTTTTT' == rc(rc('TTTTTT'))
	assert 'ACGT' == rc(rc('ACGT'))

	assert_array_almost_equal(f('TTGACC'), rc(rc(f('TTGACC'))))
	assert_array_almost_equal(f('TTTTTT'), rc(rc(f('TTTTTT'))))
	assert_array_almost_equal(f('ACGT'), rc(rc(f('ACGT'))))


def test_reverse_complement_Ns():
	assert 'TTGACCN' == reverse_complement('NGGTCAA')
	assert 'TTTNTTT' == reverse_complement('AAANAAA')
	assert 'NNN' == reverse_complement('NNN')

	f = one_hot_encode
	assert_array_almost_equal(f('TTGACCN'), reverse_complement(f('NGGTCAA')))
	assert_array_almost_equal(f('TTTNTTT'), reverse_complement(f('AAANAAA')))
	assert_array_almost_equal(f('NNN'), reverse_complement(f('NNN')))


def test_reverse_complement_disallow_Ns():
	assert_raises(ValueError, reverse_complement, "TTTGAN", allow_N=False)
	assert_raises(ValueError, reverse_complement, "N", allow_N=False)
	assert_raises(ValueError, reverse_complement, "NNNN", allow_N=False)


def test_reverse_complement_raises():
	assert_raises(ValueError, reverse_complement, "ZZ")
	assert_raises(ValueError, reverse_complement, "aCGT")


###


def test_random_one_hot():
	ohe = random_one_hot((2, 4, 9), random_state=0)

	assert ohe.shape == (2, 4, 9)
	assert ohe.dtype == torch.int8
	assert ohe.sum(axis=-2).max() == 1
	assert ohe.sum(axis=-2).min() == 1
	assert ohe.sum() == 18
	assert_array_almost_equal(ohe, [
		[[1, 0, 0, 1, 0, 0, 0, 0, 0],
         [0, 0, 1, 0, 0, 0, 0, 0, 1],
         [0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 1, 0, 0, 1, 1, 1, 1, 0]],

        [[0, 0, 0, 1, 0, 0, 1, 1, 1],
         [0, 1, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 1, 0, 0, 1, 0, 0, 0],
         [1, 0, 0, 0, 1, 0, 0, 0, 0]]])


	ohe = random_one_hot((10, 20, 15), random_state=0)

	assert ohe.shape == (10, 20, 15)
	assert ohe.dtype == torch.int8
	assert ohe.sum(axis=-2).max() == 1
	assert ohe.sum(axis=-2).min() == 1
	assert ohe.sum() == 150


def test_random_one_hot_raises_type():
	assert_raises(ValueError, random_one_hot, [0, 3, 2])
	assert_raises(ValueError, random_one_hot, 3)


def test_random_one_hot_raises_shape():
	assert_raises(ValueError, random_one_hot, (2, 3))
	assert_raises(ValueError, random_one_hot, (3,))


###


def test_chunk():
	X0 = [torch.randn(4, 10), torch.randn(4, 14)]
	X = chunk(X0, size=2)

	assert X.dtype == X0[0].dtype
	assert X.shape == (12, 4, 2)
	assert_array_almost_equal(X, torch.stack([X0[0][:, :2], X0[0][:, 2:4], 
		X0[0][:, 4:6], X0[0][:, 6:8], X0[0][:, 8:10], X0[1][:, :2], 
		X0[1][:, 2:4], X0[1][:, 4:6], X0[1][:, 6:8], X0[1][:, 8:10],
		X0[1][:, 10:12], X0[1][:, 12:14]]))

	X0 = [torch.randn(4, 10), torch.randn(4, 14)]
	X = chunk(X0, size=5)

	assert X.dtype == X0[0].dtype
	assert X.shape == (4, 4, 5)


def test_chunk_overlap():
	X0 = [torch.randn(4, 10), torch.randn(4, 14)]

	X = chunk(X0, size=2, overlap=1)
	assert X.shape == (22, 4, 2)
	assert_array_almost_equal(X[:10], torch.stack([X0[0][:, :2], X0[0][:, 1:3], 
		X0[0][:, 2:4], X0[0][:, 3:5], X0[0][:, 4:6], X0[0][:, 5:7], 
		X0[0][:, 6:8], X0[0][:, 7:9], X0[0][:, 8:10], X0[1][:, :2]]))


	X = chunk(X0, size=5, overlap=2)
	assert X.shape == (6, 4, 5)
	assert_array_almost_equal(X, torch.stack([X0[0][:, :5], X0[0][:, 3:8], 
		X0[1][:, :5], X0[1][:, 3:8], X0[1][:, 6:11], X0[1][:, 9:]]))


def test_chunk_raises():
	assert_raises(ValueError, chunk, [torch.randn(4, 10)], -1)
	assert_raises(ValueError, chunk, [torch.randn(4, 10)], 1.2)

	assert_raises(ValueError, chunk, [torch.randn(4, 10)], 2, -1)
	assert_raises(ValueError, chunk, [torch.randn(4, 10)], 2, 0.3)

	assert_raises(ValueError, chunk, torch.randn(4, 10), 2)
	assert_raises(ValueError, chunk, torch.randn(3, 4, 10), 2)


###


def test_unchunk():
	lengths = [50, 67]
	X0 = [torch.randn(4, lengths[0]), torch.randn(4, lengths[1])]

	X = chunk(X0, size=4, overlap=3)
	X1 = unchunk(X, lengths, overlap=3)
	assert_array_almost_equal(X1[0], X0[0])
	assert_array_almost_equal(X1[1], X0[1])

	X = chunk(X0, size=4, overlap=2)
	X1 = unchunk(X, lengths, overlap=2)
	assert_array_almost_equal(X1[0], X0[0])
	assert_array_almost_equal(X1[1], X0[1][:, :66])

	X = chunk(X0, size=23, overlap=8)
	X1 = unchunk(X, lengths, overlap=8)
	assert_array_almost_equal(X1[0], X0[0][:, :38])
	assert_array_almost_equal(X1[1], X0[1][:, :53])

	X = chunk(X0, size=23, overlap=0)
	X1 = unchunk(X, lengths, overlap=0)
	assert_array_almost_equal(X1[0], X0[0][:, :46])
	assert_array_almost_equal(X1[1], X0[1][:, :46])

	X0 = X0[:1]
	X = chunk(X0, size=23, overlap=6)
	X1 = unchunk(X, lengths[:1], overlap=6)
	assert_array_almost_equal(X1[0], X0[0][:, :40])


def test_unchunk_raises():
	lengths = [50, 67]
	X0 = [torch.randn(4, lengths[0]), torch.randn(4, lengths[1])]
	X = chunk(X0[:1], size=7, overlap=2)

	assert_raises(IndexError, unchunk, X, lengths, overlap=2)
	assert_raises(ValueError, unchunk, X[0], lengths[:1], overlap=2)


###


def test_extract_signal():
	loci = pandas.DataFrame({
		'example_idxs': [0, 0, 1, 2],
		'start': [1, 9, 2, 4],
		'end': [8, 14, 10, 12]
	})

	X = numpy.random.RandomState(0).randn(3, 1, 15)
	X = torch.from_numpy(X)

	y = extract_signal(loci, X)
	
	assert y.shape == (4, 1)
	assert_array_almost_equal(y, [[ 5.3088  ],
                  [ 2.891628],
                  [-0.253532],
                  [-0.917093]])

	assert y[0, 0] == X[0, :, 1:8].sum()
	assert y[1, 0] == X[0, :, 9:14].sum()
	assert y[2, 0] == X[1, :, 2:10].sum()
	assert y[3, 0] == X[2, :, 4:12].sum()



# seqlet.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# adapted from code written by Avanti Shrikumar 

import math
import numpy
import numba
import torch
import pandas

from .utils import _validate_input
from .utils import characters

from sklearn.isotonic import IsotonicRegression


def _laplacian_null(X_sum, num_to_samp=10000, random_state=1234):
	"""An internal function for calculating a null distribution.

	The TF-MoDISCo seqlet calling procedure works by constructing a null
	distribution of values using Laplacian distributions for the positive and
	negative values separately. This method constructs those Laplacian
	distributions and then samples a given number of values from them.


	Parameters
	----------
	X_sum: torch.Tensor, shape=(-1, length)
		A tensor of summed attribution values over the provided window size.

	num_to_samp: int
		The number of values to sample from the fit Laplacian distributions.

	random_state: int, optional
		The random state to use for the sampling. Default is 1234.


	Returns
	-------
	pos_values: torch.Tensor, shape=(-1,)
		A vector of sampled values from the Laplacian distribution fit to the
		positive values.

	neg_values: torch.Tensor, shape=(-1,)
		A vector of sampled values from the Laplacian distribution fit to the
		negative values.
	"""

	values = X_sum.flatten()

	hist, bin_edges = torch.histogram(values, bins=1000)
	peak = torch.argmax(hist)
	l_edge, r_edge = bin_edges[peak:peak+2]
	top_values = values[(l_edge < values) & (values < r_edge)]

	hist, bin_edges = torch.histogram(top_values, bins=1000)
	peak = torch.argmax(hist)
	mu = (bin_edges[peak] + bin_edges[peak+1]).item() / 2

	pos_values = values[values >= mu]
	neg_values = values[values <=mu]

	#Take the most aggressive lambda over all percentiles
	quantiles = torch.arange(19) * 5. / 100
	lq = -torch.log(1 - quantiles)
	pos_q = torch.quantile(a=pos_values, q=quantiles)
	neg_q = torch.quantile(a=neg_values, q=1-quantiles)

	pos_lambda = torch.max(lq / (pos_q - mu))
	neg_lambda = torch.max(lq / (torch.abs(neg_q - mu)))

	prob_pos = len(pos_values) / len(values) 
	urand = torch.from_numpy(numpy.random.RandomState(random_state).uniform(
		size=(num_to_samp, 2)))

	icdf = numpy.log(1 - urand[:, 1])

	sampled_vals = torch.where(urand[:, 0] < prob_pos, -icdf / pos_lambda + mu, 
		mu + icdf / neg_lambda)
	return sampled_vals[sampled_vals >= 0], sampled_vals[sampled_vals < 0]


def _iterative_extract_seqlets(X_sum, window_size, flank, suppress):
	"""This internal function iteratively identifies seqlets.

	This function extracts seqlets from the signal by iteratively identifying 
	the position across all examples with maximum windowed attribution sum and
	zeroing out the attributions around that position to ensure another seqlet
	cannot be identified from the same position.


	Parameters
	----------
	X_sum: torch.Tensor, shape=(-1, length)
		A tensor of summed attribution values over the provided window size.

	window_size: int
		The size of the window of attribution values to sum over when
		identifying seqlets. This is not the only component of seqlet size but
		is the most important.

	flank: int
		A number of characters on either end of the window to add to each
		seqlet. This is done primarily to remove the effect of surrounding
		positions and not have overlapping seqlets.

	suppress: int
		The number of positions to the left and right of the maximum attribution
		position to zero out the attributions of.


	Returns
	-------
	seqlets: list
		A list of tuples containing the example index, the start of the seqlet,
		the end of the seqlet, and the sum of attributions within the seqlet.
	"""

	n, d = X_sum.shape
	seqlets = []
	for i in range(n):
		while True:
			argmax = numpy.argmax(X_sum[i], axis=0)
			max_val = X_sum[i, argmax]
			if max_val == -numpy.inf:
				break

			seqlet = i, argmax - flank, argmax + window_size + flank
			seqlets.append(seqlet)

			l_idx = int(max(numpy.floor(argmax + 0.5 - suppress), 0))
			r_idx = int(min(numpy.ceil(argmax + 0.5 + suppress), d))
			X_sum[i, l_idx:r_idx] = -numpy.inf 

	return seqlets


def _isotonic_thresholds(values, null_values, increasing, target_fdr, 
	min_frac_neg=0.95):
	"""This function uses an isotonic regression to find FDR thresholds.

	Given a set of attribution values summed over the provided window and a
	set of sampled negative values from the Laplacian distribution, find the
	threshold at which real attributions are separated from the null values at
	a given FDR threshold.

	As an implementation note, this method uses the scikit-learn
	IsotonicRegression function. This method seems to accept PyTorch tensors
	as input but these tensors will be moved over to the CPU to make sure
	they are compatible. Even if other operations are done on the GPU, the
	isotonic regression component must be done on the CPU.


	Parameters
	----------
	values: torch.Tensor, shape=(-1,)
		Sampled attribution window sums from the real data.

	null_values: torch.Tensor, shape=(-1,)
		Sampled values from the Laplacian distribution that represents the null
		data.

	increasing: bool
		Whether the data should be modeled as increasing or decreasing in size.

	target_fdr: float
		The FDR threshold to use to separate positive and negative attribution
		values.

	min_frac_neg: float, optional
		The minimum number of values that need to be assigned to the null
		distribution. Default is 0.95.


	Returns
	-------
	threshold: float
		A single value representing the threshold to use on attribution window
		sums to separate seqlets from background given the FDR threshold.
	"""

	n1, n2 = len(values), len(null_values)

	X = torch.cat([values, null_values], axis=0).cpu()
	y = torch.cat([torch.ones(n1), torch.zeros(n2)], axis=0)

	w = len(values) / len(null_values)
	sample_weight = torch.cat([torch.ones(n1), torch.ones(n2)*w], axis=0)

	model = IsotonicRegression(out_of_bounds='clip', increasing=increasing)
	model.fit(X, y, sample_weight=sample_weight)
	y_hat = torch.from_numpy(model.transform(values))


	min_prec_x = model.X_min_ if increasing else model.X_max_
	min_precision = max(model.transform([min_prec_x])[0], 1e-7)
	implied_frac_neg = -1 / (1 - (1 / min_precision))

	if (implied_frac_neg > 1.0 or implied_frac_neg < min_frac_neg):
		implied_frac_neg = max(min(1.0,implied_frac_neg), min_frac_neg)

	precisions = torch.minimum(torch.maximum(1 + implied_frac_neg*(
		1 - (1 / torch.maximum(y_hat, torch.tensor(1e-7)))), torch.tensor(0.0)), 
			torch.tensor(1.0))
	precisions[-1] = 1
	return values[precisions >= (1 - target_fdr)][0].item()


def tfmodisco_seqlets(X_attr, window_size=21, flank=10, target_fdr=0.2, 
	min_passing_frac=0.03, max_passing_frac=0.2, 
	weak_threshold_for_counting_sign=0.8):
	"""Extract seqlets using the procedure from TF-MoDISco.

	Seqlets are contiguous spans of high attribution characters. This method
	for identifying them is the one that is implemented in the TF-MoDISco
	algorithm. Importantly, TF-MoDISco does several post-processing steps
	on these seqlets that are interleaved in the pattern identification 
	procedure so the final set of seqlets actually used by patterns in 
	TF-MoDISco will be smaller than the set that are returned here.

	The seqlets returned by this procedure have been optimized to be useful
	for motif discovery, and so are generally much longer and less sensitive
	than one might initially expect. The seqlets are longer because the local
	context that patterns occur in might be useful, and because uninformative
	characters on the flanks can easily be trimmed off. The seqlets are also
	less sensitive, in the sense that sometimes spans that one might call a
	seqlet by eye are missed, to prevent noise from contaminating the found
	patterns.


	Parameters
	----------
	X_attr: torch.Tensor, shape=(-1, length)
		A tensor of attribution values for each position in the sequence.
		The attributions here will be summed across the length of the alphabet
		so the values must be amenable to that. This means that, most likely,
		it should be attribution values multiplied by the one-hot encodings
		so only the present characters have attributions.

	window_size: int, optional
		The size of the window of attribution values to sum over when
		identifying seqlets. This is not the only component of seqlet size but
		is the most important. Default is 21.

	flank: int, optional
		A number of characters on either end of the window to add to each
		seqlet. This is done primarily to remove the effect of surrounding
		positions and not have overlapping seqlets. Default is 10.

	target_fdr: float, optional
		A FDR value to set on attribution score sums over windows when 
		separating called seqlets from background. Default is 0.2.

	min_passing_frac: float, optional
		Require that at least this proportion of windows pass seqlet 
		identification. Default is 0.03.

	max_passing_frac: float, optional 
		Require that no more than this proportion of windows pass seqlet
		identification. Default is 0.2.

	weak_threshold_for_counting_sign: float, optional
		A minimal threshold to use when setting the final threshold value
		separating seqlets from non-seqlets.


	Returns
	-------
	seqlets: pandas.DataFrame, shape=(-1, 4)
		A tensor containing the example index, start position, end position,
		and attribution sum for each seqlet that passes the thresholds.
	"""

	_validate_input(X_attr, "X_attr", shape=(-1, -1)) 
	suppress = int(0.5*window_size) + flank

	X_sum = X_attr.unfold(-1, window_size, 1).sum(dim=-1)
	values = X_sum.flatten()
	if len(values) > 1000000:
		values = torch.from_numpy(numpy.random.RandomState(1234).choice(
			a=values, size=1000000, replace=False))
	
	pos_values = values[values >= 0]
	neg_values = values[values < 0]
	pos_null_values, neg_null_values = _laplacian_null(X_sum)

	pos_threshold = _isotonic_thresholds(pos_values, pos_null_values, 
		increasing=True, target_fdr=target_fdr)
	neg_threshold = _isotonic_thresholds(neg_values, neg_null_values,
		increasing=False, target_fdr=target_fdr)

	values = torch.cat([pos_values, neg_values])
	frac_passing = (sum(values >= pos_threshold) + 
		sum(values <= neg_threshold)) / len(values)

	if frac_passing < min_passing_frac:
		pos_threshold = torch.quantile(torch.abs(values), q=1-min_passing_frac) 
		neg_threshold = -pos_threshold

	if frac_passing > max_passing_frac:
		pos_threshold = torch.quantile(torch.abs(values), q=1-max_passing_frac) 
		neg_threshold = -pos_threshold

	distribution = torch.sort(torch.abs(X_sum.flatten()))[0]

	transformed_pos_threshold = numpy.sign(pos_threshold) * numpy.searchsorted(
		distribution, v=abs(pos_threshold)) / len(distribution)
	transformed_neg_threshold = numpy.sign(neg_threshold) * numpy.searchsorted(
		distribution, v=abs(neg_threshold)) / len(distribution)

	idxs = (X_sum >= pos_threshold) | (X_sum <= neg_threshold)

	X_sum[idxs] = numpy.abs(X_sum[idxs])
	X_sum[~idxs] = -numpy.inf

	if flank > 0:
		X_sum[:, :flank] = -numpy.inf
		X_sum[:, -flank:] = -numpy.inf

	seqlets = _iterative_extract_seqlets(X_sum=X_sum, window_size=window_size,
		flank=flank, suppress=suppress)

	#find the weakest transformed threshold used across all tasks
	weak_thresh = min(min(transformed_pos_threshold, 
		abs(transformed_neg_threshold)) - 0.0001, 
			weak_threshold_for_counting_sign)

	threshold = distribution[int(weak_thresh * len(distribution))]

	seqlets_ = []
	for example_id, start, end, in seqlets:
		attr_flank = int(0.5 * ((end-start) - window_size))
		attr_start, attr_end = start + attr_flank, end - attr_flank
		attr = X_attr[example_id, attr_start:attr_end].sum(dim=-1).item()

		seqlet = example_id, start.item(), end.item(), attr
		seqlets_.append(seqlet)

	names = 'example_idx', 'start', 'end', 'attribution'
	return pandas.DataFrame(seqlets_, columns=names)


###


@numba.njit
def _recursive_seqlets(X, threshold=0.01, min_seqlet_len=4, max_seqlet_len=25, 
	additional_flanks=0):
	"""An internal function implementing the recursive seqlet algorithm."""

	n, l = X.shape

	# Convert observed attributions into cumsums for fast span calculation
	X_csum = numpy.zeros((n+1, l))
	for i in range(n):
		for j in range(l):
			X_csum[i, j+1] = X_csum[i, j] + X[i, j]

	xmins = numpy.empty(max_seqlet_len+1, dtype=numpy.float64)
	xmaxs = numpy.empty(max_seqlet_len+1, dtype=numpy.float64)
	X_cdfs = numpy.zeros((2, max_seqlet_len+1, 1000), dtype=numpy.float64)

	# Construct background distributions
	for j in range(min_seqlet_len, max_seqlet_len+1):
		xmin, xmax = 0.0, 0.0
		n_pos, n_neg = 0.0, 0.0

		# For a given span size, find the minimum and maximum values and the count
		# of the number of attribution values in each.
		for i in range(n):
			for k in range(l-j+1):
				x_ = X_csum[i, k+j] - X_csum[i, k]

				if x_ > 0:
					xmax = max(x_, xmax)
					n_pos += 1.0
				else:
					xmin = min(x_, xmin)
					n_neg += 1.0

		xmins[j] = xmin
		xmaxs[j] = xmax
		p_pos, p_neg = 1 / n_pos, 1 / n_neg

		# Now go through again and bin the attribution value, recording the count of
		# the number of occurences, pre-divided by the total count to get probs.
		for i in range(n):
			for k in range(l-j+1):
				x_ = X_csum[i, k+j] - X_csum[i, k]

				if x_ > 0:
					x_int = math.floor(999 * x_ / xmax)
					X_cdfs[0, j, x_int] += p_pos
				else:
					x_int = math.floor(999 * x_ / xmin)
					X_cdfs[1, j, x_int] += p_neg
					

		# Convert these PDFs into 1 - CDFs.
		for i in range(1, 1001):
			if i < 1000:
				X_cdfs[0, j, i] += X_cdfs[0, j, i-1]
				X_cdfs[1, j, i] += X_cdfs[1, j, i-1]
			
			X_cdfs[0, j, i-1] = 1 - X_cdfs[0, j, i-1]
			X_cdfs[1, j, i-1] = 1 - X_cdfs[1, j, i-1]

	###

	p_value = numpy.ones((max_seqlet_len+1, l), dtype=numpy.float64)
	seqlets = []

	# Calculate p-values for each seqlet span and keep only the maximum p-value of
	# spans starting here thus-far. Because of the recursive property, if a span
	# has a high p-value, definitionally none of the other spans including it can.
	for i in range(n):
		for j in range(min_seqlet_len, max_seqlet_len+1):
			for k in range(l-j+1):
				x_ = X_csum[i, k+j] - X_csum[i, k]

				if x_ > 0:
					x_int = math.floor(999 * x_ / xmaxs[j])
					p_value[j, k] = X_cdfs[0, j, x_int]
				else:
					x_int = math.floor(999 * x_ / xmins[j])
					p_value[j, k] = X_cdfs[1, j, x_int]
				
				if j > min_seqlet_len:
					if p_value[j-1, k] >= threshold:
						p_value[j, k] = threshold
					#p_value[j, k] = max(p_value[j-1, k], p_value[j, k])

		# Iteratively identify spans, from longest to shortest, that satisfy the
		# recursive p-value threshold.
		for j in range(max_seqlet_len - min_seqlet_len + 1):
			j = max_seqlet_len - j

			while True:
				start = p_value[j].argmin()
				p = p_value[j, start]
				p_value[j, start] = 1

				if p >= threshold:
					break

				for k in range(j - min_seqlet_len):
					if p_value[j-k, start+k+1] >= threshold:
						break

				else:
					for end in range(start, min(start+j, l-1)):
						p_value[:, end] = 1

					end = min(start + j + additional_flanks, l-1)
					start = max(start - additional_flanks, 0)
					attr = X_csum[i, end] - X_csum[i, start]
					seqlets.append((i, start, end, attr, p))

	return seqlets
		
	

def recursive_seqlets(X, threshold=0.01, min_seqlet_len=4, max_seqlet_len=25, 
	additional_flanks=0):
	"""A seqlet caller implementing the recursive seqlet algorithm.

	This algorithm identifies spans of high attribution characters, called
	seqlets, using a simple approach derived from the Tomtom/FIMO algorithms.
	First, distributions of attribution sums are created for all potential
	seqlet lengths by discretizing the sum, with one set of distributions for
	positive attribution values and one for negative attribution values. Then,
	CDFs are calculated for each distribution (or, more specifically, 1-CDFs).
	Finally, p-values are calculated via lookup to these 1-CDFs for all
	potential CDFs, yielding a (n_positions, n_lengths) matrix of p-values.

	This algorithm then identifies seqlets by defining them to have a key
	property: all internal spans of a seqlet must also have been called a
	seqlet. This means that all spans from `min_seqlet_len` to `max_seqlet_len`,
	starting at any position in the seqlet, and fully contained by the borders,
	must have a p-value below the threshold. Functionally, this means finding
	entries where the upper left triangle rooted in it is comprised entirely of
	values below the threshold. Graphically, for a candidate seqlet starting at
	X and ending at Y to be called a seqlet, all the values within the bounds
	(in addition to X) must also have a p-value below the threshold.


							min_seqlet_len
                             --------
	. . . . . . . | . . . . / . . . . . . . .
	. . . . . . . | . . . / . . . . . . . . .
	. . . . . . . | . . / . . . . . . . . . .
	. . . . . . . | . / . . . . . . . . . . .
	. . . . . . . | / . . . . . . . . . . . .
	. . . . . . . X . . . . . . . . Y . . . .
	. . . . . . . . . . . . . . . . . . . . .
	. . . . . . . . . . . . . . . . . . . . .

	
	The seqlets identified by this approach will usually be much smaller than
	those identified by the TF-MoDISco approach, including sometimes missing
	important characters on the flanks. You can set `additional_flanks` to 
	a higher value if you want to include additional positions on either side.
	Importantly, the initial seqlet calls cannot overlap, but these additional
	characters are not considered when making that determination. This means
	that seqlets may appear to overlap when `additional_flanks` is set to a
	higher value.


	Parameters
	----------
	X: torch.Tensor or numpy.ndarray, shape=(-1, length)
		Attributions for each position in each example. The identity of the
		characters is not relevant for seqlet calling, so this should be the
		"projected" attributions, i.e., the attribution of the observed
		characters.

	threshold: float, optional
		The p-value threshold for calling seqlets. All positions within the
		triangle (as detailed above) must be below this threshold. Default is
		0.01.

	min_seqlet_len: int, optional
		The minimum length that a seqlet must be, and the minimal length of
		span that must be identified as a seqlet in the recursive property.
		Default is 4.

	max_seqlet_len: int, optional
		The maximum length that a seqlet can be. Default is 25.

	additional_flanks: int, optional
		An additional value to subtract from the start, and to add to the end,
		of all called seqlets. Does not affect the called seqlets.


	Returns
	-------
	seqlets: pandas.DataFrame, shape=(-1, 5)
		A BED-formatted dataframe containing the called seqlets, ranked from
		lowest p-value to higher p-value. The returned p-value is the p-value
		of the (location, length) span and is not influenced by the other
		values within the triangle. 
	"""

	if isinstance(X, torch.Tensor):
		X = X.numpy()
	elif not isinstance(X, numpy.ndarray):
		raise ValueError("`X` must be either a torch.Tensor or numpy.ndarray.")


	columns = ['example_idx', 'start', 'end', 'attribution', 'p-value']
	seqlets = _recursive_seqlets(X, threshold, min_seqlet_len, max_seqlet_len, 
		additional_flanks)
	seqlets = pandas.DataFrame(seqlets, columns=columns)
	return seqlets.sort_values("p-value").reset_index(drop=True)
    


# test_predict.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest

from tangermeme.utils import one_hot_encode
from tangermeme.utils import random_one_hot

from tangermeme.predict import predict

from .toy_models import SumModel
from .toy_models import FlattenDense
from .toy_models import Conv
from .toy_models import Scatter
from .toy_models import ConvDense

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal


torch.manual_seed(0)
torch.use_deterministic_algorithms(True, warn_only=True)


@pytest.fixture
def X():
	return random_one_hot((64, 4, 100), random_state=0).type(torch.float32)


@pytest.fixture
def alpha():
	r = numpy.random.RandomState(0)
	return torch.from_numpy(r.randn(64, 1)).type(torch.float32)

@pytest.fixture
def beta():
	r = numpy.random.RandomState(1)
	return torch.from_numpy(r.randn(64, 1)).type(torch.float32)


##


def test_predict_summodel(X):
	torch.manual_seed(0)
	model = SumModel()
	y = predict(model, X, batch_size=8, device='cpu')

	assert y.shape == (64, 4)
	assert y.dtype == torch.float32
	assert y.sum() == X.sum()
	assert_array_almost_equal(y[:8], [
		[25., 24., 19., 32.],
        [26., 18., 29., 27.],
        [28., 25., 21., 26.],
        [21., 33., 19., 27.],
        [27., 22., 23., 28.],
        [31., 27., 21., 21.],
        [26., 20., 21., 33.],
        [21., 28., 31., 20.]])

	assert_array_almost_equal(y, model(X))
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'))
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'))


def test_predict_flattendense(X):
	torch.manual_seed(0)
	model = FlattenDense()
	y = predict(model, X, batch_size=8, device='cpu')

	assert y.shape == (64, 3)
	assert y.dtype == torch.float32

	assert_array_almost_equal(y[:8], [
		[ 0.1928,  0.2788, -0.1000],
        [-0.0505,  0.0174, -0.1751],
        [-0.1408,  0.0105,  0.0673],
        [-0.2375, -0.0069,  0.0835],
        [ 0.0442, -0.0692, -0.1565],
        [-0.3489, -0.0876, -0.2994],
        [-0.3894, -0.1332, -0.3717],
        [-0.3682,  0.5173, -0.4089]], 4)

	assert_array_almost_equal(y, model(X).detach())	
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'))
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'))


def test_predict_conv(X):
	torch.manual_seed(0)
	model = Conv()
	y = predict(model, X, batch_size=8, device='cpu')

	assert y.shape == (64, 12, 98)
	assert y.dtype == torch.float32

	assert_array_almost_equal(y[:2, :4, :10], [
		[[-0.2713, -0.5317, -0.3737, -0.4055, -0.3269, -0.3269, -0.1928,
          -0.3509, -0.4816, -0.3197],
         [-0.2988,  0.0980, -0.1013, -0.2560,  0.2595,  0.2595,  0.2167,
           0.4330, -0.0124,  0.3218],
         [-0.1247,  0.1433, -0.3111, -0.3220, -0.4084, -0.4084, -0.2111,
          -0.3884, -0.3460, -0.3052],
         [-0.1362, -0.0067,  0.5554,  0.1810,  0.2007,  0.2007, -0.1166,
           0.0337,  0.1722, -0.3441]],

        [[-0.4805, -0.1669, -0.5863, -0.0537, -0.2702, -0.1669, -0.4055,
          -0.5078, -0.0848, -0.5863],
         [-0.3709, -0.3077, -0.5910,  0.0165, -0.6573, -0.3077, -0.2560,
          -0.0755,  0.1277, -0.5910],
         [ 0.4396, -0.1558,  0.2097, -0.0929,  0.6608, -0.1558, -0.3220,
           0.1234, -0.1761,  0.2097],
         [-0.0027,  0.4644,  0.1406, -0.1111, -0.3111,  0.4644,  0.1810,
           0.1603,  0.2667,  0.1406]]], 4)

	assert_array_almost_equal(y, model(X).detach())	
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'))
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'))


def test_predict_scatter(X):
	torch.manual_seed(0)
	model = Scatter()
	y = predict(model, X, batch_size=8, device='cpu')

	assert y.shape == (64, 100, 4)
	assert y.dtype == torch.float32
	assert y.sum() == X.sum()
	assert (y.sum(axis=-1) == X.sum(axis=1)).all()
	assert (y.sum(axis=1) == X.sum(axis=-1)).all()

	assert_array_almost_equal(y[:4, :5], [
		[[1., 0., 0., 0.],
         [0., 0., 0., 1.],
         [0., 1., 0., 0.],
         [1., 0., 0., 0.],
         [0., 0., 0., 1.]],

        [[0., 1., 0., 0.],
         [0., 0., 1., 0.],
         [1., 0., 0., 0.],
         [0., 0., 0., 1.],
         [1., 0., 0., 0.]],

        [[0., 1., 0., 0.],
         [0., 0., 1., 0.],
         [0., 0., 0., 1.],
         [0., 0., 1., 0.],
         [0., 0., 1., 0.]],

        [[0., 1., 0., 0.],
         [1., 0., 0., 0.],
         [0., 0., 0., 1.],
         [1., 0., 0., 0.],
         [0., 0., 1., 0.]]], 4)

	assert_array_almost_equal(y, model(X).detach())	
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'))
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'))


def test_predict_convdense(X):
	torch.manual_seed(0)
	model = ConvDense()
	y = predict(model, X, batch_size=2, device='cpu')

	assert len(y) == 2
	assert y[0].shape == (64, 12, 98)
	assert y[1].shape == (64, 3)

	assert y[0].dtype == torch.float32
	assert y[1].dtype == torch.float32

	assert_array_almost_equal(y[0][:2, :4, :4], [
		[[-0.7757,  0.0397, -0.8799, -1.0194],
         [ 0.0947,  0.2042, -0.2302, -0.3144],
         [ 0.3781,  0.5009,  0.8585,  0.6738],
         [-0.5986, -0.0296, -0.0995, -0.2718]],

        [[-0.6676, -0.8873, -0.8696, -0.4684],
         [-0.0768, -0.0089,  0.0244,  0.1355],
         [ 0.4828,  0.5419,  0.3170,  0.3288],
         [-0.1026, -0.4676, -0.3080, -0.2606]]], 4)

	assert_array_almost_equal(y[1][:4], [
		[ 0.1928,  0.2788, -0.1000],
        [-0.0505,  0.0174, -0.1751],
        [-0.1408,  0.0105,  0.0673],
        [-0.2375, -0.0069,  0.0835]], 4)


def test_predict_convdense_batch_size(X):
	torch.manual_seed(0)
	model = ConvDense()
	y = predict(model, X, batch_size=64, device='cpu')

	assert len(y) == 2
	assert y[0].shape == (64, 12, 98)
	assert y[1].shape == (64, 3)

	assert y[0].dtype == torch.float32
	assert y[1].dtype == torch.float32


def test_predict_batch_size(X):
	torch.manual_seed(0)
	model = Scatter()
	y = predict(model, X, batch_size=68, device='cpu')
	assert y.shape == (64, 100, 4)


def test_predict_raises_shape(X):
	torch.manual_seed(0)
	model = Scatter()
	assert_raises(RuntimeError, predict, model, X[0], device='cpu')
	assert_raises(RuntimeError, predict, model, X[:, 0], device='cpu')
	assert_raises(RuntimeError, predict, model, X.unsqueeze(0), device='cpu')


def test_predict_raises_args(X, alpha, beta):
	torch.manual_seed(0)
	model = FlattenDense()
	assert_raises(TypeError, predict, model, X, batch_size=2, args=5, 
		device='cpu')
	assert_raises(AttributeError, predict, model, X, batch_size=2, args=(5,), 
		device='cpu')
	assert_raises(ValueError, predict, model, X, batch_size=2, 
		args=alpha, device='cpu')
	assert_raises(ValueError, predict, model, X, batch_size=2, 
		args=(alpha[:5],), device='cpu')
	assert_raises(ValueError, predict, model, X, batch_size=2, 
		args=(alpha, beta[:5]), device='cpu')


def test_predict_flattendense_16bit_str(X):
	torch.manual_seed(0)
	model = FlattenDense()
	y = predict(model, X, batch_size=8, dtype='float16', device='cpu')

	assert y.shape == (64, 3)
	assert y.dtype == torch.float16
	assert next(model.parameters()).dtype == torch.float32

	assert_array_almost_equal(y[:8], [
		[ 0.1928,  0.2788, -0.1000],
        [-0.0505,  0.0174, -0.1751],
        [-0.1408,  0.0105,  0.0673],
        [-0.2375, -0.0069,  0.0835],
        [ 0.0442, -0.0692, -0.1565],
        [-0.3489, -0.0876, -0.2994],
        [-0.3894, -0.1332, -0.3717],
        [-0.3682,  0.5173, -0.4089]], 3)

	assert_array_almost_equal(y, model(X).detach(), 3)
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'), 3)
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'), 3)


def test_predict_flattendense_16bit(X):
	torch.manual_seed(0)
	model = FlattenDense()
	y = predict(model, X, batch_size=8, dtype=torch.float16, device='cpu')

	assert y.shape == (64, 3)
	assert y.dtype == torch.float16
	assert next(model.parameters()).dtype == torch.float32

	assert_array_almost_equal(y[:8], [
		[ 0.1928,  0.2788, -0.1000],
        [-0.0505,  0.0174, -0.1751],
        [-0.1408,  0.0105,  0.0673],
        [-0.2375, -0.0069,  0.0835],
        [ 0.0442, -0.0692, -0.1565],
        [-0.3489, -0.0876, -0.2994],
        [-0.3894, -0.1332, -0.3717],
        [-0.3682,  0.5173, -0.4089]], 3)

	assert_array_almost_equal(y, model(X).detach(), 3)
	assert_array_almost_equal(y, predict(model, X, batch_size=1, device='cpu'), 3)
	assert_array_almost_equal(y, predict(model, X, batch_size=64, device='cpu'), 3)


# marginalize.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch

from .utils import one_hot_encode
from .ersatz import substitute
from .predict import predict


def marginalize(model, X, motif, start=None, alphabet=['A', 'C', 'G', 'T'], 
	func=predict, additional_func_kwargs={}, **kwargs):
	"""Apply a function before and after substituting a motif into sequences.

	A marginalization experiment is one where a function is applied before
	and after substituting something into a set of sequences. It is named as 
	such because the sequences are meant to be background sequences and
	difference in output before and after the substitution represent the
	"marginal" effect of adding that something into the sequences. When you are
	adding a motif to the sequence, the difference in output can be interpreted 
	as the effect that motif has on the function in isolation.

	By default, `marginalize` will apply the `predict` function to `X` before
	and after substituting in a one-hot encoded version of `motif`. However,
	one can pass in any function, including `deep_lift_shap` or even
	`saturated_mutagenesis`. These functions may have additional arguments
	and those can be passed into `marginalize` as-is and will be passed along
	to the function. If any arguments would have had the same name as those
	used by this function, you can use the `additional_func_kwargs` input to
	ensure those values get to the function.

	Naturally, most models being used with tangermeme will be non-linear and
	so the marginal effect of each motif is only somewhat useful because motifs
	do not occur in isolation in the genome. Other functions, such as `space`,
	can be invaluable in seeing how motifs interact with each other. However,
	looking at the marginal effect of each motif can still be invaluable
	because it gives you a sense for what motifs yield an effect at all and
	roughly how strong that effect is.
	

	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences to have a motif inserted into.

	motif: torch.tensor, shape=(-1, len(alphabet), motif_length)
		A one-hot encoded version of a short motif to insert into the set of
		sequences.

	start: int or None, optional
		The starting position of where to insert the motif. If None, insert the
		motif into the middle of the sequence such that the middle of the motif
		occurs at the middle of the sequence. Default is None.

	alphabet : set or tuple or list, optional
		A pre-defined alphabet where the ordering of the symbols is the same
		as the index into the returned tensor, i.e., for the alphabet ['A', 'B']
		the returned tensor will have a 1 at index 0 if the character was 'A'.
		Characters outside the alphabet are ignored and none of the indexes are
		set to 1. This is not necessary or used if a one-hot encoded tensor is
		provided for the motif. Default is ['A', 'C', 'G', 'T'].

	func: function, optional
		A function to apply before and after making the substitution. Default 
		is `predict`.

	additional_func_kwargs: dict, optional
		Additional named arguments to pass into the function when it is called.
		This is provided as an alternate path to route arguments into the 
		function in case they overlap, name-wise, with those in this function,
		or if you want to be absolutely sure that the arguments are making
		their way into the function. Default is {}.

	kwargs: optional
		Additional named arguments that will get passed into the function when
		it is called. Default is no arguments are passed in.


	Returns
	-------
	y_before: torch.Tensor or list of torch.Tensors
		The output from the function before inserting the motif in. If the
		output is a single tensor, it will return that. If the model outputs a 
		list of tensors, it will return those.

	y_after: torch.Tensor or list of torch.Tensors
		The output from the function after inserting the motif in. If the
		output from the model's forward function is a single tensor, it will
		return that. If the model outputs a list of tensors, it will return
		those.
	"""

	additional_func_kwargs = additional_func_kwargs or {}

	X_perturb = substitute(X, motif, start=start, alphabet=alphabet)
	y_before = func(model, X, **kwargs, **additional_func_kwargs)
	y_after = func(model, X_perturb, **kwargs, **additional_func_kwargs)

	return y_before, y_after


def marginalize_annotations(model, X, X0, annotations, **kwargs):
	"""Perform marginalizations on each annotation individually.

	This function takes in a model, a set of sequences, a set of background
	sequences, and a set of annotations, and returns the marginalization values
	for each annotation. For each annotation, the sequence in `X` is extracted
	and substituted into `X0` with predictions returned for `X0` before and
	after the substitution is performed, similar to the `saturation_mutagenesis`
	function. Each marginalization is done individually.

	This function will extract the sequence in each annotation and perform a
	marginalization on it individually. 


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences corresponding to the annotations.

	X0: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences that motifs will be substituted into.

	annotations: torch.Tensor, shape=(n_annotations, 3)
		A tensor of annotations where the first column is the example_idx, the
		second column is the start position (0-indexed) and the third column is
		the end position (0-indexed, not inclusive).

	kwargs: arguments
		Additional optional arguments to pass into the `ablate` function.

	
	Returns
	-------
	y_befores: torch.Tensor or list of torch.Tensors
		The application of `func` from the model BEFORE inserting the motif. If 
		the output from the model's forward function is a single tensor, it will 
		return that. If the model outputs a list of tensors, it will return 
		those.

	y_afters: torch.Tensor or list of torch.Tensors
		The application of `func` from the model AFTER inserting the motif. If 
		the output from the model's forward function is a single tensor, it will
		return that. If the model outputs a list of tensors, it will return
		those.
	"""

	y_befores, y_afters = [], []

	for idx, start, end in annotations:
		seq = X[idx, :, start:end].unsqueeze(0)

		y_before, y_after = marginalize(model, X0, seq, **kwargs)
		y_befores.append(y_before)
		y_afters.append(y_after)

	if isinstance(y_afters[0], torch.Tensor):
		y_befores = torch.stack(y_befores)
		y_afters = torch.stack(y_afters)
	else:
		y_befores = [torch.stack([x[i] for x in y_befores]) for i in range(len(
			y_befores))]
		y_afters = [torch.stack([x[i] for x in y_afters]) for i in range(len(
			y_afters))]

	return y_befores, y_afters
	